import functools
import logging
import time
from datetime import datetime

import openai
from deepchecks_llm_client.api import API
from deepchecks_llm_client.utils import handle_exceptions, set_verbosity

logging.basicConfig()
logger = logging.getLogger(__name__)


class OpenAIInstrumentor:

    def __init__(self, api: API, verbose: bool = False, log_level: int = logging.WARNING, auto_collect: bool = False):
        self.api = api
        self._auto_collect = auto_collect
        logger.setLevel(log_level)
        set_verbosity(verbose, logger)

        self.openai_version = "0.0.0"
        try:
            from importlib import metadata
            self.openai_version = metadata.version('openai')
        except Exception as ex:
            pass

    @staticmethod
    def _patched_call(original_fn, patched_fn):
        @functools.wraps(original_fn)
        def _inner_patch(*args, **kwargs):
            return patched_fn(original_fn, *args, **kwargs)

        return _inner_patch

    def auto_collect(self, auto_collect: bool):
        self._auto_collect = auto_collect
        return self

    def patcher_create(self, original_fn, *args, **kwargs):
        if self._auto_collect:
            self._before_run_log_print(args, kwargs, original_fn)

            started_at = datetime.utcnow()
            result = original_fn(*args, **kwargs)
            finished_at = datetime.utcnow()

            self._after_run_actions(args, kwargs, original_fn, result, started_at, finished_at)
        else:
            logger.debug(f"auto collect is off, skipping logging {original_fn.__qualname__} to Deepchecks")
            result = original_fn(*args, **kwargs)
        return result

    async def patcher_acreate(self, original_fn, *args, **kwargs):
        if self._auto_collect:
            self._before_run_log_print(args, kwargs, original_fn)

            started_at = datetime.utcnow()
            result = await original_fn(*args, **kwargs)
            finished_at = datetime.utcnow()

            self._after_run_actions(args, kwargs, original_fn, result, started_at, finished_at)
        else:
            logger.debug(f"auto collect is off, skipping logging {original_fn.__qualname__} to Deepchecks")
            result = await original_fn(*args, **kwargs)
        return result

    @handle_exceptions(logger)
    def _after_run_actions(self, args, kwargs, original_fn, result, started_at, finished_at):
        time_delta = (finished_at - started_at).total_seconds() * 1000
        logger.info("Finished running function: %s, time delta: %sms", original_fn.__qualname__, time_delta)
        logger.debug("Function Output: %s", result)

        # Obfuscate the api-key
        if kwargs.get("api_key"):
            kwargs["api_key"] = f"last-4-digits-{kwargs['api_key'][-4:]}"

        event_dict = {
            "request": {"func_name": original_fn.__qualname__, "args": args, "kwargs": kwargs},
            "output": result.to_dict_recursive(),
            "runtime_data": {"response_time": time_delta,
                             "openai_version": self.openai_version,
                             'started_at': started_at.isoformat(),
                             'finished_at': finished_at.isoformat()},
        }
        self.api.load_openai_data(data=[event_dict])

        logger.debug("Reported this event to deepchecks server:\n%s", event_dict)

    @staticmethod
    def _before_run_log_print(args, kwargs, original_fn):
        logger.info("Running the original function: %s. args: %s, kwargs: %s", original_fn.__qualname__, args, kwargs)

    def perform_patch(self):
        try:
            openai.ChatCompletion.acreate = self._patched_call(
                openai.ChatCompletion.acreate, self.patcher_acreate
            )
        except AttributeError:
            pass

        try:
            openai.ChatCompletion.create = self._patched_call(
                openai.ChatCompletion.create, self.patcher_create
            )
        except AttributeError:
            pass

        try:
            openai.Completion.acreate = self._patched_call(
                openai.Completion.acreate, self.patcher_acreate
            )
        except AttributeError:
            pass

        try:
            openai.Completion.create = self._patched_call(
                openai.Completion.create, self.patcher_create
            )
        except AttributeError:
            pass



import logging
import typing as t
import uuid
from datetime import datetime
from io import StringIO

import pandas as pd

from deepchecks_llm_client.api import API
from deepchecks_llm_client.data_types import Tag, EnvType, AnnotationType, GoldenSetInteraction, Step
from deepchecks_llm_client.openai_instrumentor import OpenAIInstrumentor
from deepchecks_llm_client.utils import handle_exceptions, set_verbosity, handle_generator_exceptions

__all__ = ["dc_client", "DeepchecksLLMClient"]


logging.basicConfig()
logger = logging.getLogger(__name__)
init_logger = logging.Logger(__name__ + ".init")

DEFAULT_VERSION_NAME = '0.0.1'
DEFAULT_ENV_TYPE = EnvType.PROD


class DeepchecksLLMClientError(Exception):
    pass


class DeepchecksLLMClient:
    def __init__(self):
        self._api: API = None
        self.app: t.Dict[str, t.Any] = None
        self.instrumentor: OpenAIInstrumentor = None
        self._initialized: bool = False
        self._verbose: bool = False
        self._log_level: int = logging.WARNING

    @handle_exceptions(init_logger)
    def init(self,
             host: str,
             api_token: str,
             app_name: str,
             version_name: str = DEFAULT_VERSION_NAME,
             env_type: EnvType = DEFAULT_ENV_TYPE,
             auto_collect: bool = True,
             init_verbose: bool = True,
             verbose: bool = False,
             log_level: int = logging.WARNING
             ):
        """
        Connect to Deepchecks LLM Server

        Parameters
        ==========
        host : str
            Deepchecks host to communicate with
        api_token : str
            Deepchecks API Token (can be generated from the UI)
        app_name : str
            Application name to connect to
        version_name : str, default='0.0.1'
            Version name to connect to inside the application,
            if Version name does not exist SDK will create it automatically,
        env_type : EnvType, default=EnvType.PROD
            could be EnvType.PROD (for 'Production') or EnvType.EVAL (for 'Evaluation')
        auto_collect : bool, default=True
            Auto collect calls to LLM Models
        init_verbose: bool, default=True
            Write log messages during the init phase and non-prod phases such as golden_set()
        verbose : bool, default=False
            Write log messages (by default we are in silence mode)
        log_level: int, default=logging.WARNING
            In case that verbose or init_verbose is True,
            this parameter will set SDK loggers logging level

        Returns
        =======
        None
        """
        if self._initialized:
            logger.warning(
                "Deepchecks client was initialized already. "
                "We will ignore newly provided parameters"
            )
            return

        logger.setLevel(log_level)
        self._log_level = log_level
        set_verbosity(init_verbose, init_logger)
        set_verbosity(verbose, logger)
        self._verbose = verbose

        if self._api is None:
            if host is not None and api_token is not None:
                self._api = API.instantiate(host=host, token=api_token)
            else:
                raise DeepchecksLLMClientError('host/token parameters must be provided')


        if app_name is None:
            raise DeepchecksLLMClientError('app_name must be supplied')
        self.app = self._api.get_application(app_name)
        if not self.app:
            raise DeepchecksLLMClientError(f'Application: "{app_name}", does not exist, please create it via the UI')

        self.app_name(app_name).version_name(version_name).env_type(env_type)

        self.instrumentor = None
        if auto_collect:
            self.instrumentor = OpenAIInstrumentor(self.api, verbose, log_level, auto_collect)
            self.instrumentor.perform_patch()
        self._initialized = True

    @property
    def initialized(self):
        return self._initialized

    @property
    @handle_exceptions(logger)
    def api(self) -> API:
        if self._api:
            return self._api
        raise DeepchecksLLMClientError("dc_client was not initialized correctly, please re-create it")

    @handle_exceptions(logger, return_self=True)
    def app_name(self, new_app_name: str):
        self.api.app_name(new_app_name)
        return self

    @handle_exceptions(logger, return_self=True)
    def version_name(self, new_version_name: str):
        self.api.version_name(new_version_name)
        return self

    @handle_exceptions(logger, return_self=True)
    def env_type(self, new_env_type: EnvType):
        self.api.env_type(new_env_type)
        return self

    @handle_exceptions(logger)
    def auto_collect(self, enabled: bool):
        if not self.instrumentor:
            self.instrumentor = OpenAIInstrumentor(
                api=self.api, verbose=self._verbose, log_level=self._log_level, auto_collect=enabled,
            )
            self.instrumentor.perform_patch()
        else:
            self.instrumentor.auto_collect(auto_collect=enabled)
        return self

    @handle_exceptions(logger)
    def set_tags(self, tags: t.Dict[Tag, str]):
        self.api.set_tags(tags)
        return self

    @handle_exceptions(logger)
    def annotate(self, user_interaction_id: str, version_name: str,
                 annotation: AnnotationType = None, reason: t.Optional[str] = None):
        """
        Annotate a specific interaction by its user_interaction_id
        Parameters
        ----------
        user_interaction_id
            Unique id of the interaction, can be set by the user or automatically generated by deepchecks
        version_name
            Version name to annotate
        annotation
            Could be one of AnnotationType.GOOD, AnnotationType.BAD, AnnotationType.UNKNOWN
            or None to remove annotation
        reason
            String that explains the reason for the annotation

        Returns
        -------
            None

        """

        version_obj = self._get_version_object(version_name)
        self.api.annotate(user_interaction_id, version_obj["id"], annotation, reason=reason)

    @handle_exceptions(logger)
    def log_interaction(self,
                        input: str,
                        output: str,
                        full_prompt: str = None,
                        information_retrieval: str = None,
                        annotation: AnnotationType = None,
                        annotation_reason: str = None,
                        user_interaction_id: str = None,
                        started_at: t.Union[datetime, None] = None,
                        finished_at: t.Union[datetime, None] = None,
                        steps: t.List[Step] = None,
                        custom_props: t.Dict[str, t.Any] = None) -> t.Union[str, uuid.UUID, None]:
        result = self.api.log_interaction(input=input,
                                          output=output,
                                          full_prompt=full_prompt,
                                          information_retrieval=information_retrieval,
                                          annotation=annotation,
                                          annotation_reason=annotation_reason,
                                          user_interaction_id=user_interaction_id,
                                          started_at=started_at.isoformat() if started_at else None,
                                          finished_at=finished_at.isoformat() if finished_at else None,
                                          steps=steps,
                                          custom_props=custom_props)
        return result.json()

    @handle_exceptions(logger)
    def update_application_config(self, file):
        self.api.update_application_config(application_id=self.app['id'], file=file)

    @handle_exceptions(logger)
    def get_application_config(self, file_save_path: t.Union[str, None] = None) -> str:
        return self.api.get_application_config(application_id=self.app['id'], file_save_path=file_save_path)

    @handle_generator_exceptions(init_logger)
    def golden_set_iterator(self, version_name: t.Union[str, None] = None) -> t.Iterable[GoldenSetInteraction]:
        """
        Fetch all interactions from the golden set (EnvType.EVAL), as iterable
        Supports pagination, so this API is suitable for iterating large amount of data

        Parameters
        ----------
        version_name : str
            version name to fetch interactions from, if no version name was supplied
            the latest created version in the application will be used (mostly useful
            when integrating the golden set and rerun it in the CI)

        Returns
        -------
        Iterable collection of interactions

        """
        golden_set_version = self._get_version_object(version_name)

        offset = 0
        limit = 20

        while True:
            interactions = self.api.get_interactions(golden_set_version["id"], limit=limit, offset=offset)
            for interaction in interactions:
                yield GoldenSetInteraction(user_interaction_id=interaction["user_interaction_id"],
                                           input=interaction["input"]["data"] if interaction.get("input") else None,
                                           information_retrieval=interaction["information_retrieval"]["data"] if interaction.get("information_retrieval") else None,
                                           full_prompt=interaction["prompt"]["data"] if interaction.get("prompt") else None,
                                           output=interaction["output"]["data"] if interaction.get("output") else None,
                                           created_at=interaction["created_at"],
                                           topic=interaction["topic"],
                                           output_properties=interaction["output_properties"] if interaction.get("output_properties") else [])

            # If the size of the data is less than the limit, we've reached the end
            if len(interactions) < limit:
                break

            offset += limit

    @handle_exceptions(init_logger)
    def golden_set(self, version_name: t.Union[str, None] = None, return_topics: bool = False,
                   return_output_props: bool = False, return_annotation_data: bool = False) -> t.Union[pd.DataFrame, None]:
        """
        Fetch all the interactions from the golden set (EnvType.EVAL) as pandas dataframe

        Parameters
        ----------
        version_name : str
            version name to fetch interactions from, if no version name was supplied
            the latest created version in the application will be used (mostly useful
            when integrating the golden set rerun in the CI)

        return_topics : bool
            include topic in the data

        return_output_props : bool
            include output properties in the data

        return_annotation_data : bool
            include annotation info in the data

        Returns
        -------
        Pandas dataframe or None in case of a problem retrieving the data

        """
        golden_set_version = self._get_version_object(version_name)
        csv_as_text = self.api.get_interactions_csv(golden_set_version["id"],
                                                    return_topics=return_topics,
                                                    return_output_props=return_output_props,
                                                    return_annotation_data=return_annotation_data)
        data = StringIO(csv_as_text)
        return pd.read_csv(data)

    @handle_exceptions(init_logger)
    def golden_set_generation_completed(self):
        """
        This function indicate deepchecks that golden set generation completed
        It helps to finalize golden set related actions, for instance: train topics
        model and extract topics (the model will be used later on to extract topics
        from production data)

        Returns
        -------
        None

        """
        self.api.generate_topics()

    def _get_version_object(self, version_name):

        app = self.api.get_application(self.api.get_app_name())
        if not app:
            raise DeepchecksLLMClientError(f"Application: '{self.api.get_app_name()}', does not exist")
        if version_name:
            golden_set_version = next((ver for ver in app["versions"] if ver["name"] == version_name), None)
            if not golden_set_version:
                raise DeepchecksLLMClientError(f"Could not find version '{version_name}', in application '{self.api.get_app_name()}'")
        else:
            golden_set_version = max(app["versions"], key=lambda x: x['created_at'])
            if not golden_set_version:
                raise DeepchecksLLMClientError(f"Could not find versions to select from in application '{self.api.get_app_name()}'")
        return golden_set_version


# LLM Client publicly accessed singleton
dc_client: DeepchecksLLMClient = DeepchecksLLMClient()



# -----------------------------------------------------------------------------
# Deepchecks LLM Onboarding Script
# -----------------------------------------------------------------------------
# This script is designed to guide you through the process of integrating
# data with Deepchecks LLM Product, using Deepchecks' SDK.
# Please follow the instructions provided and replace placeholders with actual
# data as needed.


# Step 1: Import necessary modules
# Create new venv and install deepchecks' SDK using `pip install deepchecks-llm-client`
import uuid
from datetime import datetime

import openai

from deepchecks_llm_client.client import dc_client
from deepchecks_llm_client.data_types import Tag, EnvType, AnnotationType, Step, StepType


def main():
    # Step 2: Set up configuration

    # Fill deepchecks host name here
    DEEPCHECKS_LLM_HOST = "https://host-name-here"  # could be: https://app.llm.deepchecks.com

    # Login to deepchecks' service and generate new API Key (Configuration -> API Key) and place it here
    DEEPCHECKS_LLM_API_KEY = "Fill Key Here"

    # Use "Update Samples" in deepchecks' service, to create a new application name and place it here
    DEEPCHECKS_APP_NAME = "DemoApp"

    # In case you would like to experience with deepchecks auto collect capabilities, please set
    # Open AI's key here (as string)
    OPENAI_API_KEY = "Fill Key Here"

    # Step 3: Initialize the SDK

    # auto_collect=True wraps `openai.ChatCompletion` and `openai.Completion` APIs
    # so any OpenAI invocation will fire an event to Deepchecks with the relevant data
    # If you do not plan to use OpenAI auto collection, you can set it to `False`

    # Notice - deepchecks SDK was designed to be non-intrusive. It does not throw exceptions
    # and only print to log. By default, it prints only errors from the "init" phase
    # If you wish to increase verbosity, use `verbose=True` and `log_level=logging.INFO`
    version = "0.0.1"
    env_type = EnvType.EVAL

    print(f"init deepchecks client, host: {DEEPCHECKS_LLM_HOST}, app: {DEEPCHECKS_APP_NAME}, version: {version}, env: {env_type}")
    dc_client.init(host=DEEPCHECKS_LLM_HOST,
                   api_token=DEEPCHECKS_LLM_API_KEY,
                   app_name=DEEPCHECKS_APP_NAME,
                   version_name=version,
                   env_type=env_type,
                   auto_collect=True)

    # Step 4: Auto collecting OpenAI Data (Q&A + Annotation)
    if OPENAI_API_KEY:
        # You can set_tags() to add additional information to the collected data
        # most important is the INPUT, which is the actual user question, without this
        # information, deepchecks won't be able to evaluate the question the user asked
        INPUT = "How much is 2 plus 2?"
        # If you have "information retrieval" step in your flow, you can supply is as well
        # using the Tag.INFORMATION_RETRIEVAL tag
        INFORMATION_RETRIEVAL = "1 plus 1 equals 2"
        # Notice that you can add USER_ID (optional) to track the number of unique users that are
        # using your LLM application (remember PII and obfuscate the username if needed)
        dc_client.set_tags({Tag.INPUT: INPUT,
                            Tag.USER_ID: "A05fdfbb2035e@gmail.com",
                            Tag.INFORMATION_RETRIEVAL: INFORMATION_RETRIEVAL})

        # Setting up your OpenAI API credentials
        print("calling OpenAI ChatCompletion API")
        openai.api_key = OPENAI_API_KEY
        chat_completion = openai.ChatCompletion.create(model="gpt-3.5-turbo",
                                                       temperature=0.7,
                                                       messages=[
                                                       {"role": "system", "content": f"Act as a calculator {INFORMATION_RETRIEVAL}"},
                                                       {"role": "user", "content": INPUT}])

        # The data has already logged to deepchecks server, you can annotate the data
        # in anytime you like using OpenAI's unique id
        # Notice that you can set your own unique id using Tags.USER_INTERACTION_ID (before calling
        # OpenAI's API), and use it here instead of OpenAI's id
        print(f"Annotating OpenAI interaction with key: {chat_completion.openai_id}")
        dc_client.annotate(chat_completion.openai_id, annotation=AnnotationType.GOOD)

        # clear tags so we won't carry them on to the next calls
        # relevant only for "auto collected" data
        dc_client.set_tags({})
    else:
        print("OpenAI Key was not supply, skipping OpenAI demo")

    # Step 5: Explicit Logging of data and optionally addition of steps for flow visibility
    # Notice that we can also log custom properties - you can use custom properties that were
    # already defined via the UI, or use the "CP_xxx" convention to create new custom property
    # on the fly

    # We can explicitly log interaction
    print(f"log interaction using explicit Deepchecks API: log_interaction()")
    user_inter_id = str(uuid.uuid4())
    dc_client.log_interaction(input="my user input",
                          output="my model output",
                          full_prompt="system part: my user input",
                          information_retrieval="system part: my information retrieval",
                          annotation=AnnotationType.BAD,
                          user_interaction_id=user_inter_id,
                          started_at=datetime(2021, 10, 31, 15, 1, 0).astimezone(),
                          finished_at=datetime.utcnow().astimezone(),
                          steps=[
                              Step(
                                  name="Information Retrieval",
                                  type=StepType.INFORMATION_RETRIEVAL,
                                  attributes={"env": "dev"}),
                              Step(
                                  input="my user input + the retried data",
                                  name="Information Retrieval",
                                  type=StepType.LLM,
                                  attributes={"env": "dev"},
                                  annotation=AnnotationType.BAD,
                                  output="my model output"),
                              Step(
                                  name="Information Retrieval",
                                  type=StepType.TRANSFORMATION,
                                  attributes={"env": "dev"},
                                  output="my model output after PII removal")
                          ],
                          custom_props={"CP_MY_PROP1": 0.1})

    # Notice that we can re-annotate our logged interation in a later phase
    # using out own external id
    print(f"Annotating explicit interaction with key: {user_inter_id}")
    dc_client.annotate(user_inter_id, AnnotationType.GOOD)

    # Indicate we finished process golden set data for 0.0.1
    dc_client.golden_set_generation_completed()

    # Step 6: Creating a new version based on the "0.0.1" golden set

    # Set different version / env type, to be carried on to the next calls
    # Usually, in "dev mode", you will log to "EnvType.EVAL" and use a different version
    # for each evaluation you make. Check the performance of your LLM application using deepchecks,
    # and deploy the selected version to production, and from that point on,
    # you will use "EnvType.PROD" with the "production version" to log production data
    new_version = "0.0.2"
    eval_env_type = EnvType.EVAL
    print(f"switching to version: {new_version}, and EnvType: {eval_env_type}")
    dc_client.version_name(new_version).env_type(eval_env_type)

    # Now we have 1 or 2 interactions logged into deepchecks' under version '0.0.1'
        # We can get the "golden set" out of deepchecks and rerun it against OpenAI while log the
    # new result to our new version (i.e. - "0.0.2")
    if OPENAI_API_KEY:

        # golden_set_iterator() can get "version_name" as param, if no version_name was supplied
        # the latest created version will be taken (this is useful when integrating it in the CI)
        for interaction in dc_client.golden_set_iterator(version_name="0.0.1"):
            print(f"processing golden set interaction with unique_interaction_id: {interaction.user_interaction_id}  "
                  f"(created_at: {interaction.created_at})")
            dc_client.set_tags({Tag.INPUT: interaction.input,
                                Tag.INFORMATION_RETRIEVAL: interaction.information_retrieval})

            # We run the same OpenAI command, only changing the temperature...
            chat_completion = openai.ChatCompletion.create(model="gpt-3.5-turbo",
                                                           api_key=OPENAI_API_KEY,
                                                           temperature=0.2,
                                                           messages=[
                                                           {"role": "system", "content": f"Act as a calculator {interaction.information_retrieval}"},
                                                           {"role": "user", "content": interaction.input}])

            # Indicate we finished process golden set data for 0.0.2
            dc_client.golden_set_generation_completed()

    # One can also retrieve golden set as dataframe, for example:
    df = dc_client.golden_set(version_name="0.0.1")
    print("We can also fetch the golden set as dataframe, here are the dataframe columns")
    print(list(df.columns))


if __name__ == '__main__':
    main()
